#Import required Gurobi package
#--------------------------------------------------------------------------------------------------------------------
import gurobipy as gp
import pandas as pd
import numpy as np
# Define the function that is callable from R and returns the distribution vector
# that prescribes to which the patient, that currently needs to be transferred, is transferred to
#--------------------------------------------------------------------------------------------------------------------

def run_optimization_program_transferring_patient(capacity_ward, capacity_ICU, excess_shortage_current_ward,\
                                                  excess_shortage_current_icu, excess_shortage_future_ward, excess_shortage_future_icu,\
                                                  costs_transfer):

    #Parameter definitions
    #--------------------------------------------------------------------------------------------------------------------

    # Hieronder voegen we het buitenland toe en bepalen we positieve/negatieve deel van de beschikbare capaciteit vectoren
    BigM = 10000000
    nscen = np.shape(excess_shortage_future_icu)[0]


    capacity_ward.append(BigM)  # add the foreign country to the capacity vector (has infinite ward capacity)
    capacity_ICU.append(BigM)   # add the foreign country to the capacity vector (has infinite ICU capacity)

    excess_shortage_future_ward = np.append(excess_shortage_future_ward, np.ones(shape = (nscen, 1))*BigM, axis = 1)
    excess_shortage_current_ward.append(BigM)   # add the foreign country to the expected free capacity vector (has infinite free ward capacity) 
    excess_shortage_future_icu = np.append(excess_shortage_future_icu, np.ones(shape = (nscen, 1))*BigM, axis = 1)   # add the foreign country to the expected free capacity vector (has infinite free ICU capacity)
    excess_shortage_current_icu.append(BigM)    # add the foreign country to the expected free capacity vector (has infinite free ICU capacity)

    # excess_current_ward = [max(0, x) for x in excess_shortage_current_ward]   # [ \tilde{n}_{rW}^{+}(s,t) ]
    # excess_current_icu = [max(0, x) for x in excess_shortage_current_icu]     # [ \tilde{n}_{rI}^{+}(s,t) ]

    # shortage_current_ward = [max(0, -x) for x in excess_shortage_current_ward]   # [ \tilde{n}_{rW}^{-}(s,t) ]
    # shortage_current_icu = [max(0, -x) for x in excess_shortage_current_icu]     # [ \tilde{n}_{rI}^{-}(s,t) ]

    # shortage_current_ward.append(0)    # add foreign country to binary indication vector if ward transfer is required (for the foreign country always equal to 0)
    # shortage_current_icu.append(0)     # add foreign country to binary indication vector if ICU transfer is required (for the foreign country always equal to 0)
    # Definieer de kosten voor verplaatsen van regio r naar regio k [ \gamma_{r,k}(s,t)]
    R = len(capacity_ward) - 1 # R equals number of regions (R+1 denotes the foreign country in model formulation; in Python this is R since counting from 0)

    #Definition optimization program
    #--------------------------------------------------------------------------------------------------------------------
    # Create optimization model
    model = gp.Model('ILP_rasters_poli')
    model.Params.LogToConsole = 0
    # Create flow variables
    f_here_and_now_ward = model.addVars(range(R+1), range(R+1), lb=0, vtype=gp.GRB.INTEGER, name = 'f_here_and_now_ward')
    f_here_and_now_ICU = model.addVars(range(R+1), range(R+1), lb=0, vtype=gp.GRB.INTEGER, name = 'f_here_and_now_ICU')

    f_wait_and_see_ward = model.addVars(range(nscen), range(R+1), range(R+1), lb=0, vtype=gp.GRB.INTEGER, name="f_wait_and_see_ward")
    f_wait_and_see_ICU = model.addVars(range(nscen), range(R+1), range(R+1), lb=0, vtype=gp.GRB.INTEGER, name="f_wait_and_see_ICU")

    # Create here-and-now constraints
    for r in range(R+1):
        other_regions = [k for k in range(R+1) if k != r]
        model.addConstr(max(0,excess_shortage_current_ward[r]) - f_here_and_now_ward.sum(other_regions,r) >= 0)
        model.addConstr(min(0,excess_shortage_current_ward[r]) + f_here_and_now_ward.sum(r,other_regions) == 0)
        
        model.addConstr(max(0,excess_shortage_current_icu[r]) - f_here_and_now_ICU.sum(other_regions,r)>= 0)
        model.addConstr(min(0,excess_shortage_current_icu[r]) + f_here_and_now_ICU.sum(r,other_regions) == 0)
        
        # model.addConstr(excess_shortage_current_ward[r] + f_here_and_now_ward.sum(r,other_regions)  - f_here_and_now_ward.sum(other_regions,r)  >= 0)
        # model.addConstr(excess_shortage_current_icu[r] + f_here_and_now_ICU.sum(r,other_regions)  - f_here_and_now_ICU.sum(other_regions,r)  >= 0)
        
    #here-and-now flow to other country should be equal to the total shortage in the country
    model.addConstr(f_here_and_now_ward.sum(other_regions,R) + min(0, sum(excess_shortage_current_ward[:R]))==0)
    model.addConstr(f_here_and_now_ICU.sum(other_regions,R) + min(0, sum(excess_shortage_current_icu[:R]))==0)

    # Create wait-and-see constraints
    for s in range(nscen):
        for r in range(R+1):
            other_regions = [k for k in range(R+1) if k != r]
    
            model.addConstr(excess_shortage_future_ward[s,r] + f_here_and_now_ward.sum(r,other_regions)  - f_here_and_now_ward.sum(other_regions,r) +
                            f_wait_and_see_ward.sum(s,r,other_regions)  - f_wait_and_see_ward.sum(s,other_regions,r) >= 0)
            
            model.addConstr(excess_shortage_future_icu[s,r] + f_here_and_now_ICU.sum(r,other_regions)  - f_here_and_now_ICU.sum(other_regions,r) +
                            f_wait_and_see_ICU.sum(s,r,other_regions)  - f_wait_and_see_ICU.sum(s,other_regions,r) >= 0)



    ### Define relative remaining excess capacity (variables in the optimization program; denoted with delta) [ \delta_{r,W}(s), \delta_{r,W}(s,t) ]
    delta_here_and_now_ward = model.addVars(range(R), vtype=gp.GRB.CONTINUOUS, lb=0, name="delta_here_and_now_ward")
    delta_here_and_now_ICU = model.addVars(range(R), vtype=gp.GRB.CONTINUOUS, lb=0, name="delta_here_and_now_ICU")
    delta_wait_and_see_ward = model.addVars(range(nscen),range(R), vtype=gp.GRB.CONTINUOUS, lb=0, name="delta_wait_and_see_ward")
    delta_wait_and_see_ICU = model.addVars(range(nscen),range(R), vtype=gp.GRB.CONTINUOUS, lb=0, name="delta_wait_and_see_ICU")

    for r in range(R):
        other_regions = [k for k in range(R+1) if k != r]

        model.addConstr(delta_here_and_now_ward[r] == \
                        (excess_shortage_current_ward[r] + f_here_and_now_ward.sum(r,other_regions) - f_here_and_now_ward.sum(other_regions, r))/capacity_ward[r])

        model.addConstr(delta_here_and_now_ICU[r] == \
                        (excess_shortage_current_icu[r] + f_here_and_now_ICU.sum(r,other_regions) - f_here_and_now_ICU.sum(other_regions, r))/capacity_ICU[r])
        
        for s in range(nscen):
            model.addConstr(delta_wait_and_see_ward[s,r] == \
                            (excess_shortage_future_ward[s,r] + f_here_and_now_ward.sum(r,other_regions) - f_here_and_now_ward.sum(other_regions, r) + \
                              f_wait_and_see_ward.sum(s,r,other_regions) - f_wait_and_see_ward.sum(s,other_regions, r))/capacity_ward[r])
    
            model.addConstr(delta_wait_and_see_ICU[s,r] == \
                           (excess_shortage_future_icu[s,r] + f_here_and_now_ICU.sum(r,other_regions) - f_here_and_now_ICU.sum(other_regions, r)+\
                            f_wait_and_see_ICU.sum(s,r,other_regions) - f_wait_and_see_ICU.sum(s,other_regions, r))/capacity_ICU[r])

    # Set objective function
    model.setObjective(sum(costs_transfer[r,k]*f_here_and_now_ward.sum(r,k) for r in range(R+1) for k in range(R+1))+ \
                   sum(costs_transfer[r,k]*f_here_and_now_ICU.sum(r,k) for r in range(R+1) for k in range(R+1))+ \
                   sum(costs_transfer[r,k]*f_wait_and_see_ward.sum(s,r,k) for s in range(nscen) for r in range(R+1) for k in range(R+1))/nscen+ \
                   sum(costs_transfer[r,k]*f_wait_and_see_ICU.sum(s, r,k) for s in range(nscen) for r in range(R+1) for k in range(R+1))/nscen + \
                   sum(delta_here_and_now_ward[r]*delta_here_and_now_ward[r] for r in range(R)) + \
                   sum(delta_here_and_now_ICU[r]*delta_here_and_now_ICU[r] for r in range(R)) + \
                   sum(delta_wait_and_see_ward[s,r]*delta_wait_and_see_ward[s,r] for s in range(nscen) for r in range(R))/nscen + \
                   sum(delta_wait_and_see_ICU[s,r]*delta_wait_and_see_ICU[s,r] for s in range(nscen) for r in range(R))/nscen, \
                   gp.GRB.MINIMIZE)
    # Solve
    model.update()
    model.Params.IntegralityFocus = 1
    model.Params.IntFeasTol = 1E-9
    model.optimize()

    f_here_and_now_ward_optimal = [[] for _ in range(R+1)]
    f_here_and_now_ICU_optimal = [[] for _ in range(R+1)]

    f_wait_and_see_ward_optimal = [[] for _ in range(R+1)]
    f_wait_and_see_ICU_optimal = [[] for _ in range(R+1)]

    for r in range(R+1):
        for k in range(R+1):
            f_here_and_now_ward_optimal[r].append(round(f_here_and_now_ward.sum(r,k).getValue()))
            f_here_and_now_ICU_optimal[r].append(round(f_here_and_now_ICU.sum(r,k).getValue()))

            # f_wait_and_see_ward_optimal[r].append(round(f_wait_and_see_ward.sum(r,k).getValue()))
            # f_wait_and_see_ICU_optimal[r].append(round(f_wait_and_see_ICU.sum(r,k).getValue()))
    
    
    return pd.DataFrame.from_dict(f_here_and_now_ward_optimal), pd.DataFrame.from_dict(f_here_and_now_ICU_optimal),\
           pd.DataFrame.from_dict(f_wait_and_see_ward_optimal), pd.DataFrame.from_dict(f_wait_and_see_ICU_optimal), model

if __name__ == '__main__':
    excess_shortage_current_ward = [11, -3, 21]
    excess_shortage_current_icu  = [-8, 10, -10]
    excess_shortage_future_ward = np.array([[-21,  -9, 13], [-21,  -9, 13],[-21,  -9, 13]])
    excess_shortage_future_icu  = np.array([[-15, -1 ,-12], [-15, -1 ,-12],[-15, -1 ,-12]])
    capacity_ward = [60, 65, 50]
    capacity_ICU  = [27, 35, 18]
    costs_transfer    = np.matrix([[1e3, 1e3, 1e3, 1e6],
                              [1e3, 1e3, 1e3, 1e6],
                              [1e3, 1e3, 1e3, 1e6],
                              [1e6,1e6,1e6,1e6]])
    
    
    res = run_optimization_program_transferring_patient(capacity_ward, capacity_ICU, excess_shortage_current_ward,\
                                                      excess_shortage_current_icu, excess_shortage_future_ward, excess_shortage_future_icu,\
                                                      costs_transfer)












